# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from django.conf import settings
from django.core.mail import EmailMessage
from django.contrib.humanize.templatetags.humanize import naturaltime
from django.template.loader import render_to_string
from django.utils import timezone
from django.db.models.aggregates import Sum

from guichet.utils.registration import registration_to_image
from guichet.models import Registration, Option, Ticket
from accounts.models import User
from paypal.standard.ipn.models import PayPalIPN
from paypal.standard.models import ST_PP_COMPLETED

from reportlab.platypus import SimpleDocTemplate, Table, TableStyle, Preformatted
from reportlab.lib.styles import getSampleStyleSheet
from reportlab.lib import colors
from unidecode import unidecode
import json
import dateutil.parser


def mail_confirm_registration(registration, force_email=None):

    user = registration.owner
    email = user.email
    ticket = registration.ticket
    event = ticket.event
    options = registration.options.all()

    if force_email:
        email = force_email

    if user.is_anonyme and force_email is None:
        return

    if registration.info_state == Registration.STATE_CAN_ENTER:
        if ticket.event.email_message:
            content = ticket.event.email_message
        elif ticket.event.description:
            content = ticket.event.description
        elif ticket.long_notes:
            content = ticket.long_notes
        elif ticket.design:
            content = "Veuillez trouver ci-joint votre place pour l’évènement « {} ».".format(event.name)
        else:
            content = "Votre réservation « {} » pour l’évènement « {} » a bien été enregistrée.".format(
                ticket.name, event.name
            )

        if options:
            content += "\n Les options que vous avez choisies sont : \n"
            for o in options:
                content += " - {} : {} \n".format(o.name, o.description)

        if ticket.design:
            subject = "Billetterie: votre billet « {} » pour l’évènement « {} ».".format(ticket.name, event.name)
        else:
            subject = "Billetterie: confirmation de votre réservation pour l’évènement « {} ».".format(event.name)

        email = EmailMessage(subject, content, settings.DEFAULT_FROM_EMAIL, [email], reply_to=[event.email_event])

        if ticket.design:
            pdf = registration_to_image(registration, format="PDF")
            email.attach("billet.pdf", pdf, "application/x-pdf")
    else:
        info_state = registration.get_state_info_sup()
        subject = "Billetterie: confirmation de votre réservation pour l’évènement « {} ».".format(event.name)
        content = render_to_string(
            "emails/paid_without_all_info.txt",
            {
                "registration": registration,
                "info_missing": info_state["missing"],
                "info_rejectd": info_state["rejected"],
                "info_pending": info_state["pending"],
            },
        )
        email = EmailMessage(subject, content, settings.DEFAULT_FROM_EMAIL, [email], reply_to=[event.email_event])

    email.send()


def get_registrations_stats(registrations):

    return {
        "total": registrations.count(),
        "paid": registrations.filter(paid=True).count(),
        "valid": registrations.filter(paid=True, opposed=False, info_state=Registration.STATE_CAN_ENTER).count(),
        "cannot_enter": registrations.filter(paid=True, opposed=False, info_state=Registration.STATE_CAN_PAID).count(),
        "opposed": registrations.filter(paid=True, opposed=True).count(),
        "refund": registrations.filter(paid=True, opposed=True, refund=True).count(),
        "not_paid": registrations.filter(paid=False).count(),
        "cannot_paid": registrations.filter(paid=False, info_state=Registration.STATE_CANNOT_PAID).count(),
        "used": registrations.filter(paid=True, opposed=False, used=True).count(),
    }


def compute_event_stats(event):

    # Donne la version des statistiques générés
    # Pour les évènements archivés, seuls le retour de cette fonction
    # est conservé. La version permet de savoir l'affichage qui doit être utilisé
    version = 3

    registrations = Registration.objects.filter(ticket__event=event)
    options = Option.objects.filter(ticket__event=event)

    tickets = []
    tickets_radar = []
    for ticket in event.tickets.all():
        if ticket.allowed_godson or ticket.allowed_external:
            if ticket.allowed_external:
                statistique = get_registrations_stats(
                    registrations.filter(ticket=ticket, payee__isnull=True, owner__login_inp_net="")
                )
                statistique.update({"name": "{} (extérieurs)".format(ticket.name)})
                tickets.append(statistique)
                tickets_radar.append(statistique)
            statistique = get_registrations_stats(
                registrations.filter(ticket=ticket, payee__isnull=True).exclude(owner__login_inp_net="")
            )
            statistique.update({"name": "{} (étudiants inp)".format(ticket.name)})
            tickets.append(statistique)
            tickets_radar.append(statistique)
            statistique = get_registrations_stats(registrations.filter(ticket=ticket, payee__isnull=False))
            statistique.update({"name": "{} (parrainages)".format(ticket.name)})
            tickets.append(statistique)
            tickets_radar.append(statistique)
        statistique = get_registrations_stats(registrations.filter(ticket=ticket))
        statistique.update({"name": ticket.name})
        tickets.append(statistique)
        if not ticket.allowed_godson and not ticket.allowed_external:
            tickets_radar.append(statistique)
    tickets.append(get_registrations_stats(registrations))
    tickets[-1].update({"name": "Tous"})
    registrations = registrations.filter(paid=True)

    opts = []
    for option in options:
        opts += [
            {
                "name": option.name,
                "group_name": option.group_name,
                "count": registrations.filter(options__id=option.id).count(),
                "ticket": option.ticket.name
            }
        ]

    methods = []
    total = 0
    fees = 0
    refunds = 0
    for method, name in Registration.PAYMENT_METHODS:
        # deux requètes pour éviter des erreurs de calcul sur price_ticket si plusieurs options dans un ticket
        price_ticket = (
            registrations.filter(payment_method=method).aggregate(price_ticket=Sum("ticket__price"))["price_ticket"]
            or 0
        )
        price_options = (
            registrations.filter(payment_method=method).aggregate(price_options=Sum("options__price"))["price_options"]
            or 0
        )
        profit = price_ticket + price_options
        refund = (
            registrations.filter(payment_method=method, refund=True).aggregate(price_refund=Sum("ticket__price"))[
                "price_refund"
            ]
            or 0
        )
        refund += (
            registrations.filter(payment_method=method, refund=True).aggregate(price_options=Sum("options__price"))[
                "price_options"
            ]
            or 0
        )

        if method == Registration.PAYMENT_PAYPAL:
            pks = registrations.filter(payment_method=method).values("pk")
            invoices = map((lambda x: settings.PAYPAL_INVOICE_PREFIX + str(x["pk"])), pks)
            fee = (
                PayPalIPN.objects.exclude(flag=True)
                .filter(payment_status=ST_PP_COMPLETED, invoice__in=invoices)
                .aggregate(fee=Sum("mc_fee"))["fee"]
                or 0
            )
        else:
            fee = 0
        fees += fee
        total += profit
        refunds += refund
        methods += [
            {
                "name": name,
                "count": registrations.filter(payment_method=method).count(),
                "profit": profit,
                "fee": fee,
                "refund": refund,
                "real_profit": profit - fee - refund,
            }
        ]
    methods += [
        {
            "name": "Total",
            "count": registrations.count(),
            "profit": total,
            "fee": fees,
            "refund": refunds,
            "real_profit": total - fees - refunds,
        }
    ]

    c = {
        "version": version,
        "tickets": tickets,
        "tickets_radar": tickets_radar,
        "options": opts,
        "methods": methods,
        "registrations": list(registrations.order_by("purchase_date").values_list("purchase_date", flat=True)),
        "used_registrations": list(
            registrations.filter(used=True).order_by("used_date").values_list("used_date", flat=True)
        ),
    }

    return c


def get_event_stats(event):

    if not event.archived:
        return compute_event_stats(event)

    c = json.loads(event.archived_stats)

    # Since version1, registrations and used_registrations is date
    c["used_registrations"] = map(dateutil.parser.parse, c["used_registrations"])
    c["registrations"] = map(dateutil.parser.parse, c["registrations"])

    return c


def validate_registration(registration):
    if not registration.paid:
        return (False, "Ce billet n’a pas été payé")
    elif registration.opposed:
        return (False, "Ce billet est en opposition")
    elif registration.used:
        string = "Le titulaire de ce billet est déjà entré à {:%Hh%M} ({})".format(
            timezone.localtime(registration.used_date), naturaltime(registration.used_date)
        )
        return (False, string)
    elif registration.info_state != Registration.STATE_CAN_ENTER:
        string = "Le titulaire de ce billet n'a pas fourni tous les documents demandés pour participer à cet évènement"
        return (False, string)
    else:
        return (True, registration)


def check_registration(event, search):

    if len(search) > 0 and search[0] == "%":
        return (False, "Billet de test de template")
    # Recherche du billet par login inp de l'entrant
    try:
        user = User.objects.get(login_inp_net=search)
        registration = user.registrations.get(ticket__event=event, payee__isnull=True)
        return validate_registration(registration)
    except User.DoesNotExist:
        return (False, "Pas de billet associés à cette recherche")
    except Registration.DoesNotExist:
        return (False, "Cet utilisateur n’a pas de billet pour cet évènement")


def check_registration_sandbox(event, search):
    # detecte les billets de tests

    if len(search) < 2 or search[0] != "%":
        return (False, "Ce billet n'est pas un billet de test")
    ident = search[1:].split("%", 1)[0]

    try:
        t = Ticket.objects.get(id=ident, event=event)
    except Ticket.DoesNotExist:
        return (False, "Pas d'event associé")
    return (True, t)


def list_registrations(event, buffer):

    style = getSampleStyleSheet()["Normal"]

    def process(reg):

        payee = reg.get_real_payee()
        ticket = reg.ticket.name
        opposed = "OPPOSITION" if reg.opposed else ("" if reg.info_state == reg.STATE_CAN_ENTER else "MANQUE DOCS")
        used = "X" if reg.used else "    "

        if reg.opposed or reg.used:
            payee += "***"

        options_list = []
        for opt in reg.options.all():
            if opt.group:
                options_list.append("{} : {}".format(opt.group_name, opt.name))
            else:
                options_list.append("{}".format(opt.name))

        # formating
        payee_format = Preformatted(payee, style, maxLineLength=23, newLineChars="  ") if payee else ""
        ticket = Preformatted(ticket, style, maxLineLength=20, newLineChars="  ") if ticket else ""
        options = (
            Preformatted("\n".join(options_list), style, maxLineLength=25, newLineChars="  ") if options_list else ""
        )

        return [payee, payee_format, ticket, options, opposed, used]

    registrations = Registration.objects.filter(ticket__event=event, paid=True)
    registrations = list(map(process, registrations))
    registrations.sort(key=lambda x: unidecode(x[0]).lower())
    registrations = list(map(lambda x: x[1:], registrations))

    style = TableStyle(
        [
            ("VALIGN", (0, 0), (-1, -1), "MIDDLE"),
            ("ALIGN", (2, 0), (2, -1), "CENTER"),
            ("TEXTCOLOR", (4, 0), (4, -1), colors.red),
            ("INNERGRID", (0, 0), (-1, -1), 0.25, colors.black),
            ("BOX", (0, 0), (-1, -1), 0.25, colors.black),
        ]
    )

    t = Table(registrations)
    t.setStyle(style)

    doc = SimpleDocTemplate(buffer)
    elements = []

    elements.append(t)

    doc.build(elements)


def maj_state_registration(registration):
    if registration.info_state == Registration.STATE_CAN_ENTER:
        return
    new_state = Registration.STATE_CAN_ENTER
    info_state = registration.get_state_info_sup()
    for info in info_state["missing"] + info_state["rejected"] + info_state["pending"]:
        if info.reference.after_payement:
            new_state = min(Registration.STATE_CAN_PAID, new_state)
        else:
            new_state = Registration.STATE_CANNOT_PAID
    if registration.info_state != new_state:
        if registration.info_state > new_state:
            print(
                "retour en arrière de l'état du billet {} : {} => {}".format(
                    registration.id, registration.info_state, new_state
                )
            )
        registration.info_state = new_state
        registration.save()
        if new_state == Registration.STATE_CAN_ENTER and registration.paid:
            mail_confirm_registration(registration)
